import itertools
import os

from tfopgen.util import (parse_args, load_config, parse_inout,
                          parse_attr_type, camel_to_snake_case,
                          header_guard, jinja_env_factory)


def make_template_kwargs(op_name, py_op_name,
                         project, library, op_inputs, op_outputs,
                         op_type_attrs, op_other_attrs, op_doc):
    """
    Creates a dictionary suitable for rendering the jinja2 templates in this package
    """

    NB = '_namespace_begin'
    NE = '_namespace_stop'

    type_constraints = [tuple(t for t in a.np_types) for a in op_type_attrs]

    # Permute the type constraints
    op_tf_type_perms = itertools.product(*(a.tf_types for a in op_type_attrs))
    op_tf_type_perms = [list(p) for p in op_tf_type_perms]

    op_np_type_perms = itertools.product(*(a.np_types for a in op_type_attrs))
    op_np_type_perms = [list(p) for p in op_np_type_perms]

    # Create dictionary with variables required for creating the templates
    template_kwargs = {
        # Names
        'op_name': op_name,
        'py_op_name': py_op_name,
        'project': project,
        'library': library,
        'kernel_name': ''.join([library, '_', py_op_name]),

        # Operator inputs, outputs, attributes and documentation
        'op_inputs': op_inputs,
        'op_outputs': op_outputs,
        'op_type_attrs': op_type_attrs,
        'op_other_attrs': op_other_attrs,
        'op_tf_type_perms': op_tf_type_perms,
        'op_np_type_perms': op_np_type_perms,
        'type_constraints': type_constraints,
        'op_doc': op_doc,

        # Filenames
        'main_header_file': ''.join([py_op_name, '_op.h']),
        'cpp_header_file': ''.join([py_op_name, '_op_cpu.h']),
        'cpp_source_file': ''.join([py_op_name, '_op_cpu.cpp']),
        'cuda_header_file': ''.join([py_op_name, '_op_gpu.cuh']),
        'cuda_source_file': ''.join([py_op_name, '_op_gpu.cu']),
        'python_test_file': ''.join(['test_', py_op_name, '.py']),
        'makefile': 'Makefile',
        'shared_library': ''.join([library, '.so']),

        # C++ namespace
        'project_namespace_start': ''.join([project, NB]).upper(),
        'project_namespace_stop': ''.join([project, NE]).upper(),
        'op_namespace_start': ''.join([project, '_', py_op_name, NB]).upper(),
        'op_namespace_stop': ''.join([project, '_', py_op_name, NE]).upper(),
    }

    template_kwargs.update({
        # C++ header guards
        'main_header_guard': header_guard(library, template_kwargs['main_header_file']),
        'cpp_header_guard': header_guard(library, template_kwargs['cpp_header_file']),
        'cuda_header_guard': header_guard(library, template_kwargs['cuda_header_file']),
    })

    return template_kwargs


def run(args):
    """
    Runs the operator generator

    Arguments:
        args: list
            List of command line arguments stripped of the program name.
            sys.argv[1:] is appropriate in most cases.
    """
    args = parse_args(args)
    cfg = load_config(args.config)

    try:
        op_name = cfg['name']
        library = cfg['library']
        project = cfg['project']
    except KeyError as e:
        raise ValueError("Key '{}' was not present in '{}'"
            .format(e.message, args.config))

    op_inputs = cfg.get('inputs', [])
    op_outputs = cfg.get('outputs', [])
    op_type_attrs = cfg.get('type_attrs', [])
    op_other_attrs = cfg.get('other_attrs', [])
    op_doc = cfg.get('doc', "Documentation")

    # Parse input ops
    op_inputs = [parse_inout(i, s) for i, s in op_inputs]

    # Parse output ops
    op_outputs = [parse_inout(o, s) for o, s in op_outputs]

    # Parse type constrained attrs
    op_type_attrs = [parse_attr_type(a) for a in op_type_attrs]

    # Snake case python version of the operator
    py_op_name = camel_to_snake_case(op_name)

    template_path = os.path.join(os.path.dirname(__file__), 'templates')
    jinja_env = jinja_env_factory(template_path)

    # Create library directory if it does not exist
    if not os.path.exists(library):
        os.makedirs(library)

    # Create dictionary for rendering jinja2 templates
    kwargs = make_template_kwargs(op_name, py_op_name,
                                  project, library, op_inputs, op_outputs,
                                  op_type_attrs, op_other_attrs, op_doc)

    def render(template, output):
        """ Hook to render template file to output """
        with open(os.path.join(library, kwargs[output]), 'w') as f:
            header_template = jinja_env.get_template(template)
            f.write(header_template.render(**kwargs))

    render('main_header.j2', 'main_header_file')
    render('cpp_header.j2', 'cpp_header_file')
    render('cpp_source.j2', 'cpp_source_file')
    render('cuda_header.j2', 'cuda_header_file')
    render('cuda_source.j2', 'cuda_source_file')
    render('test_source.j2', 'python_test_file')
    render('Makefile.j2', 'makefile')



import argparse
from collections import namedtuple
import itertools
import os
import re

import jinja2
import ruamel.yaml

FIRST_CAP_RE = re.compile('(.)([A-Z][a-z]+)')
ALL_CAP_RE = re.compile('([a-z0-9])([A-Z])')


def camel_to_snake_case(name):
    """ Convert CamelCase op names to snake case """
    s1 = FIRST_CAP_RE.sub(r'\1_\2', name)
    return ALL_CAP_RE.sub(r'\1_\2', s1).lower()


def header_guard(library, header_name):
    """ Derive a C++ header guard from the library and header name """
    guard_str = header_name.replace('.', '_')
    return ''.join([library, '_', guard_str]).upper()


def strip_and_split(s, sep):
    """ Split s on sep and strip the resulting elements """
    return (c.strip() for c in s.split(sep))


def parse_args(args):
    """ Parse and return arguments """
    default_conf_path = os.path.join(
        os.path.dirname(__file__), 'conf', 'default.yml')

    parser = argparse.ArgumentParser("Tensorflow Custom Operator Generator")
    parser.add_argument('config', help="Configuration File", default=default_conf_path)
    return parser.parse_args(args)


def load_config(config_file):
    """ Load the configuration file """
    with open(config_file, "r") as f:
        return ruamel.yaml.load(f, ruamel.yaml.RoundTripLoader)


def jinja_env_factory(template_path):
    """ Creates a jinja environment, loading templates from template_path """
    jinja_loader = jinja2.FileSystemLoader(template_path)
    jinja_env = jinja2.Environment(loader=jinja_loader,
                                   trim_blocks=False, lstrip_blocks=False)

    class LastJoiner(object):
        """
        Opposite of jinja2.utils.Joiner.
        Won't emit separator on the last iteration.
        """

        def __init__(self, sep=u', '):
            self.sep = sep

        def __call__(self, loop_last):
            if loop_last:
                return u''

            return self.sep

    jinja_env.globals['last_joiner'] = LastJoiner

    # Create a filter for formatting a list
    jinja_env.filters['format_list'] = lambda l, p: [p % s for s in l]

    return jinja_env

# Create types for Inputs/Outputs
InOut = namedtuple("InOut", ["name", "type",
                             "tf_type", "np_type", "shape"])
# Create types for Typed Attributes
Attr = namedtuple("Attr", ["original", "name", "types",
                           "tf_types", "np_types", "default"])


def parse_inout(s, shape):
    """ Parse .Input() and .Output() directives """
    var, type_ = tuple(c.strip() for c in s.split(":"))

    if "*" in type_:
        raise ValueError("Failed to parse '{}'. "
                         "List lengths are not yet supported".format(s))

    from tensorflow.python.framework.dtypes import (
        _STRING_TO_TF,
        _TYPE_TO_STRING,
        _TF_TO_NP)

    TF_TYPES = _TYPE_TO_STRING.values()
    tf_type = "tensorflow::" + type_ if type_ in TF_TYPES else type_
    np_type = ("np." + _TF_TO_NP[_STRING_TO_TF[type_]].__name__
               if type_ in _STRING_TO_TF else type_)

    # Set a default shape for variable if None exists
    shape = (1024, ) if shape is None else shape

    return InOut(var, type_, tf_type, np_type, shape)


def parse_attr_type(s):
    """
    Parse type attribute directives. For example
    "FT: {float, double} = DT_FLOAT64"
    """

    # Separate s into "FT" and "{float, double} = DT_FLOAT64"
    var, types = tuple(strip_and_split(s, ":"))

    # Separate types into "{float, double}", "DT_FLOAT64"
    split = types.split("=")
    default = split[1].strip() if len(split) > 1 else None
    types = split[0].strip()

    # Handle the multiple types case
    if types.startswith("{") and types.endswith("}"):
        types = tuple(c.strip() for c in types[1:-1].split(","))
    else:
        types = (types,)

    from tensorflow.python.framework.dtypes import (
        _STRING_TO_TF,
        _TYPE_TO_STRING,
        _TF_TO_NP)

    TF_TYPES = _TYPE_TO_STRING.values()
    tf_types = tuple("tensorflow::" + t if t in TF_TYPES else t for t in types)
    np_types = ["np." + _TF_TO_NP[_STRING_TO_TF[t]].__name__
                if t in _STRING_TO_TF
                else "np." + t for t in types]

    return Attr(s, var, types, tf_types, np_types, default)



import os
import subprocess

# Versioning code here, based on
# http://blogs.nopcode.org/brainstorm/2013/05/20/pragmatic-python-versioning-via-setuptools-and-git-tags/
def maintain_version(version_file):
    """
    Get the version from git tags, and write it to version.py.
    When git is not available (PyPi package), use version in version.py.
    """

    try:
        version_git = (subprocess
            .check_output(['git', 'describe', '--tags'])
            .rstrip()
            .decode('utf-8'))
    except:
        with open(version_file, 'r') as fh:
            version_git = (fh.read()
                            .strip()
                            .split('=')[-1]
                            .replace('"','')
                            .decode('utf-8'))

    version_msg = ("# Do not edit this file, "
        "pipeline versioning is governed by git tags")

    with open(version_file, 'w') as fh:
        components = [version_msg, os.linesep,
                        '__version__="', version_git, '"']
        print([type(c) for c in components])
        content = ''.join(c for c in components)
        fh.write(content)

    return version_git


import os

from setuptools import setup, find_packages

from install.versioning import maintain_version

def readme(filename):
  return open(filename).read()

setup(name='tfopgen',
      description='Generates tensorflow custom operator boilerplate',
      long_description=readme('README.rst'),
      url='',
      classifiers=[
          "Development Status :: 3 - Alpha",
          "Intended Audience :: Developers",
          "License :: OSI Approved :: MIT License",
          "Operating System :: OS Independent",
          "Programming Language :: Python",
          "Topic :: Software Development :: Libraries :: Python Modules",
      ],
      author='Simon Perkins',
      author_email='simon.perkins@gmail.com',
      license='',
      packages=find_packages(),
      include_package_data=True,
      install_requires=[
          'jinja2 >= 2.8.0',
          'numpy >= 1.12.0',
          'ruamel.yaml >= 0.13.7'
      ],
      scripts=[os.path.join('tfopgen', 'bin', 'tfopgen')],
      version=maintain_version(os.path.join('tfopgen', 'version.py')),
      zip_safe=True)
